import datetime
import logging
import os
import re
import sys
import traceback
import urlparse
from os.path import join, isfile

from aspen import http, gauntlet, simplates, sockets
from aspen.http.request import Request
from aspen.http.response import Response
from aspen.configuration import Configurable


log = logging.getLogger('aspen.website')


class Website(Configurable):
    """Represent a website.
    """

    def __init__(self, argv=None):
        """Takes an argv list, without the initial executable name.
        """
        self.configure(argv)
        log.info("Aspen website loaded from %s." % self.root)
    
    def __call__(self, environ, start_response):
        """WSGI interface.
        """
        request = Request.from_wsgi(environ) # too big to fail :-/
        response = self.handle(request)
        return response(environ, start_response)

    def start(self):
        log.info("Starting up Aspen website.")
        self.run_hook('startup')
        self.engine.start()

    def stop(self):
        log.info("Shutting down Aspen website.")
        self.run_hook('shutdown')
        self.engine.stop()

    def run_hook(self, name):
        self.hooks.run(name, self)

    def handle(self, request):
        """Given an Aspen request, return an Aspen response.
        """
        try:
            try:
                self.copy_configuration_to(request)
                request.website = self

                self.hooks.run('inbound_early', request)
                socket = sockets.get(request)
                request.fs = self.translate(request)
                self.hooks.run('inbound_late', request)

                simplate = simplates.get(request)
                if socket is not None:
                    response = socket.handle(request)
                else:
                    response = simplate.handle(request)
            except:
                response = self.handle_error_nicely(request)
        except Response, response:
            # Grab the response object in the case where it was raised.  In the
            # case where it was returned from simplates.handle, response is set
            # in a try block above.
            pass
        else:
            # If the response object is coming from handle_error via except 
            # Response, then it already has request on it and the early hooks
            # have already been run. If it fell off the edge un-exceptionally,
            # we need to take care of those two things.
            response.request = request
            self.hooks.run('outbound_early', response)

        self.hooks.run('outbound_late', response)
        response.headers.set('Content-Length', len(response.body))
        self.log_access(request, response) # TODO is this at the right level?
        return response

    def handle_error_nicely(self, request):
        """Try to provide some nice error handling.
        """
        try:                        # nice error messages
            tb_1 = traceback.format_exc()
            response = sys.exc_info()[1]
            if not isinstance(response, Response):
                log.error(tb_1)
                response = Response(500, tb_1)
            elif 200 <= response.code < 300:
                return response
            response.request = request
            self.hooks.run('outbound_early', response)
            fs = self.ours_or_theirs(str(response.code) + '.html')
            if fs is None:
                fs = self.ours_or_theirs('error.html')
            if fs is None:
                raise
            request.fs = fs
            response = simplates.handle(request, response)
            return response
        except Response, response:  # no nice error template available
            raise
        except:                     # last chance for tracebacks in the browser
            tb_2 = traceback.format_exc().strip()
            tbs = '\n\n'.join([tb_2, "... while handling ...", tb_1])
            log.error(tbs)
            if self.show_tracebacks:
                response = Response(500, tbs)
            else:
                response = Response(500)
            response.request = request
            raise response

    def find_ours(self, filename):
        """Given a filename, return a filepath.
        """
        return join(os.path.dirname(__file__), 'www', filename)

    def ours_or_theirs(self, filename):
        """Given a filename, return a filepath or None.
        """
        ours = self.find_ours(filename)
        theirs = join(self.root, '.aspen', filename)
        if isfile(theirs):
            out = theirs
        elif isfile(ours):
            out = ours
        else:
            out = None
        return out

    def translate(self, request):
        """Given a Request, return a filesystem path, or raise Response.
        """
       
        # First step.
        # ===========
        # Set request.fs initially, return a list of fspath parts.

        parts = gauntlet.translate(request)
        log.debug("got request for " + request.fs)


        # The Gauntlet
        # ============
        # Keep request.fs up to date for logging purposes. It is used in
        # log_access, below, which could be triggered by any of the raises
        # herein. Each of these sets request.fs, and returns None or raises.

        gauntlet.check_sanity(request)
        gauntlet.hidden_files(request)
        gauntlet.virtual_paths(request, parts)
        gauntlet.trailing_slash(request)
        gauntlet.index(request)
        gauntlet.autoindex( request
                          , self.conf.aspen.no('list_directories')
                          , self.ours_or_theirs('autoindex.html')
                           )
        gauntlet.socket_files(request)
        gauntlet.not_found(request, self.find_ours('favicon.ico'))

        
        # Now you are one of us.
        # ======================

        return request.fs

    def log_access(self, request, response):
        """Log access.
        """

        # What was the URL path translated to?
        # ====================================

        fs = request.fs[len(self.root):]
        if fs:
            fs = '.'+fs
        else:
            fs = request.fs
        log.info("%s => %s" % (request.path.raw, fs))


        # Where was response raised from?
        # ===============================

        tb = sys.exc_info()[2]
        if tb is None:
            log.info("%33s" % '<%s>' % response)
        else:
            while tb.tb_next is not None:
                tb = tb.tb_next
            frame = tb.tb_frame
            co = tb.tb_frame.f_code
            filename = tb.tb_frame.f_code.co_filename
            if filename.startswith(self.root):
                filename = '.'+filename[len(self.root):]
            log.info("%33s  %s:%d" % ( '<%s>' % response
                                     , filename
                                     , frame.f_lineno
                                      ))



"""Implement the server side of Socket.IO.

https://github.com/learnboost/socket.io-spec

"""
import uuid
from aspen import Response


TRANSPORTS = ['xhr-polling']
FFFD = u'\ufffd'.encode('utf-8')


__cache__ = {}


def get(request):
    """Takes a Request object.

    The path after *.sock is something like 1/websocket/43ef6fe7?foo=bar.

    The socket.io handshake is a GET request to 1/. After the handshake,
    subsequent messages are to 1/%transport/%sid?%query.

    """

    parts = request.path.raw.split('.sock/')
    socket = None
    if len(parts) != 1:
        return None
        request.path.raw = parts[0] + '.sock'
        socket = socket_io.get_socket(parts[1], request)


    # Parse and validate the input.
    # =============================

    parts = path.split('/')
    nparts = len(parts)
    if nparts not in (2, 3):
        msg = "Expected 2 or 3 path parts for Socket.IO socket, got %d."
        raise Response(400, msg % nparts)

    protocol = parts[0]
    if protocol != '1':
        msg = "Expected Socket.IO protocol version 1, got %s."
        raise Response(400, msg % protocol)


    # Handshake
    # =========

    if len(parts) == 2:
        sid = uuid.uuid4().hex
        heartbeat = str(15)
        timeout = str(10)
        transports = ",".join(TRANSPORTS)

        __cache__[sid] = None

        handshake = ":".join([sid, heartbeat, timeout, transports])

        response = Response(200)
        response.body = handshake
        raise response # TODO return would be faster


    # More than a handshake.
    # ======================

    transport = parts[1]
    sid = parts[2]

    if transport not in TRANSPORTS:
        msg = "Expected transport in {%s}, got %s."
        msg %= (",".join(transports), transport)
        raise Response(400, msg)

    if sid not in __cache__:
        msg = "Expected %s in cache, didn't find it"
        raise Response(400, msg % sid)

    if __cache__[sid] is None:
        # This is the first request after a handshake. It's not until this
        # point that we know what transport the client wants to use.
        SpecificSocket = XHRPollingSocket
        __cache__[sid] = SpecificSocket(sid) 

    socket = __cache__[sid]
    return socket 


class Socket(object):
    """Base class.
    """

    def __init__(self, sid):
        self.sid = sid


class XHRPollingSocket(Socket):

    def handle(self, request):
        """
        """
        if request.method == 'POST':
            message = request.socket.recv()
            request.socket.send(messages[0]) 
        else:
            while 1:
                if request.socket.messages:
                    request.socket.send("1::")
                time.sleep(0.5)

        response = Response(200)
        response.body = ""
        return response

    def recv(self):
        raw = self.request.body.raw
        frames = raw.split(FFFD)
        frames.pop() # packet starts with FFFD
        assert len(frames) % 2 == 0, "Odd number of frames!"
        messages = []
        while frames:
            nbytes = frames.pop()
            bytes = frames.pop()
            messages.append(bytes)
        return messages 

    def send(self, msg):
        response = Response(200)
        msg = msg.encode('utf-8')
        msg = '%s%d%s%s' % (FFFD, len(msg), FFFD, msg)
        response.body = msg
        raise response



from aspen.simplates import PAGE_BREAK


class Simplate(object):
    """This is a regular simplate. Override the hooks for *.json and *.sock.
    """

    def __init__(self, raw):
        raw = raw.replace("^L", PAGE_BREAK)
        pages = raw.split(PAGE_BREAK)
        npages = len(pages)
        self.one, self.two, self.three = self.parse(pages, npages)

    def parse(self, pages, npages):
        """Given a sequence of raw strings, return three page strings.
        """
        if npages == 1:
            one = two = None
            three = pages[0] 
        elif npages == 2:
            one = "" # Empty string instead of None; only None if two is None
            two, three = pages
        elif npages == 3:
            one, two, three = pages
        else:
            raise SyntaxError( "Simplate %s may have at most " % request.fs
                             + "three pages; it has %d." % npages
                              )

        
        three = self.trim_initial_newline(three)

        return (one, two, three)
   
    def trim_initial_newline(self, template):
        """Trim any initial newline from page three.
        
        This is a convenience. It's nice to put ^L on a line by itself, but
        really you want the template to start on the next line.

        """
        try:
            if template[0] == '\r':
                if template[1] == '\n':
                    template = template[2:]
            elif template[0] == '\n':
                template = template[1:]
        except IndexError:
            pass


    def handle(self, request, response=None):
        """Given a Request, return or raise a Response.
        """
        if response is None:
            response = Response()

        mimetype, namespace, script, template = load(request)

        if namespace is None:
            response.body = template
        else:
           
            # Populate namespace.
            # ===================
        
            namespace.update(request.namespace)
            namespace['request'] = request
            namespace['response'] = response
       

            # Exec the script.
            # ================
        
            if script:
                exec script in namespace
                response = namespace['response']


            # Process the template.
            # =====================
            # If template is None that means that that page was empty.
        
            if template is not None:
                response.body = template.generate(**namespace)

        
        # Set the mimetype.
        # =================
        # We guess based on the filesystem path, not the URL path. Also, we 
        # special case JSON.
       
        if mimetype == 'application/json':
            if template is None:                # dynamic
                if not isinstance(response.body, basestring):
                    # json.dumps is guaranteed to exist here.
                    response.body = json.dumps( response.body
                                              , cls=FriendlyEncoder
                                               )
            else:                               # static
                pass
            response.headers.set('Content-Type', request.json_content_type)
        
        if response.headers.one('Content-Type') is None:
            if mimetype.startswith('text/'):
                mimetype += "; charset=UTF-8" 
            response.headers.set('Content-Type', mimetype)


        # Send it on back up the line.
        # ============================

        return response




from aspen import json
from aspen.simplates.simplate import Simplate


class JSONSimplate(Simplate):

    def __init__(self, *a, **kw):
        """Extend to check for json. By now we know we're not static.
        """
        if json is None:
            raise LoadError("Neither json nor simplejson was found. Try "
                            "installing simplejson to use dynamic JSON "
                            "simplates. See "
                            "http://aspen.io/simplates/json/#libraries for "
                            "more information.")
        super(JSONSimplate, self).__init__(self, *a, **kw)

    def parse(self, pages, npages):
        """Override to look for two pages instead of three.
        """

        three  = None
        if npages == 1:
            one = None
            two = pages[0]
        elif npages == 2:
            one, two = pages
        else:
            raise SyntaxError( "JSON simplate %s may have at " % request.fs
                             + "most two pages; it has %d." % npages
                              )

        return (one, two, three) 
